#include "outputCommon.h"
#include "IOutputInfo.h"
#include <windows.h>
#include <cstring>
#include <string>
#include <map>

using namespace System;
using namespace System::IO;
using namespace System::Collections::Generic;
using namespace System::Runtime::InteropServices;
using namespace System::Windows::Forms;
using namespace System::Threading;

namespace ruche {
namespace aviutl {
namespace plugins
{
    ////////////////////
    // OUTPUT_INFO pR[obN֐fQ[g`
    // ĂяoK __cdecl

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate void* GetVideoCallback(int);

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate void* GetAudioCallback(int, int, int*);

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate BOOL IsAbortCallback();

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate BOOL RestTimeDispCallback(int, int);

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate int GetFlagCallback(int);

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate BOOL UpdatePreviewCallback();

    [UnmanagedFunctionPointer(CallingConvention::Cdecl)]
    private delegate void* GetVideoExCallback(int, DWORD);

    ////////////////////

    // QƃJE^t̃vOCf[^
    private class PluginData
    {
    public:
        // RXgN^
        PluginData(::HMODULE dllHandle, ::OUTPUT_PLUGIN_TABLE* table)
            :
            _refCount(new int(0)),
            _dllHandle(dllHandle),
            _table(table)
        {
            ref();
        }

        // Rs[RXgN^
        PluginData(const PluginData& src)
            :
            _refCount(src._refCount),
            _dllHandle(src._dllHandle),
            _table(src._table)
        {
            ref();
        }

        // Rs[Zq
        PluginData& operator=(const PluginData& r)
        {
            PluginData temp(r);
            this->unref();
            this->_refCount = temp._refCount;
            this->_dllHandle = temp._dllHandle;
            this->_table = temp._table;
            this->ref();
        }

        // fXgN^
        ~PluginData()
        {
            unref();
        }

        // vOCnh擾
        ::HMODULE getDllHandle()
        {
            return _dllHandle;
        }

        // o̓vOCe[u擾
        ::OUTPUT_PLUGIN_TABLE* getTable()
        {
            return _table;
        }

    private:
        // QƃJE^CNg
        void ref()
        {
            ++(*_refCount);
        }

        // QƃJE^fNgAQƂ؂ꂽꍇ͉s
        void unref()
        {
            if (--(*_refCount) <= 0)
            {
                if (_table != 0 && _table->func_exit != 0)
                {
                    _table->func_exit();
                }
                if (_dllHandle != 0)
                {
                    ::FreeLibrary(_dllHandle);
                }
                delete _refCount;
            }
        }

    private:
        int* _refCount;
        ::HMODULE _dllHandle;
        ::OUTPUT_PLUGIN_TABLE* _table;
    };

    // vOCf[^}bv^
    typedef std::map<std::wstring, PluginData> PluginDataMap;

    // vOCe[u
    static PluginDataMap PluginTable;

    ////////////////////

    // AviUtl̏o̓vOCbvNXB
    public ref class OutputPlugin
    {
    public:
        // ݒt@CۑfBNg̃pXB
        static initonly String^ ConfigDataDirPath;

    private:
        // W[̃tpXB
        String^ _filePath;

        // W[̃t@CB
        String^ _fileName;

        // W[nhB
        HMODULE _hPlugin;

        // vOCC^tF[XB
        OUTPUT_PLUGIN_TABLE* _pi;

        // BeginPlugin ĂяoJE^B
        int _piRefCount;

        // o̓^CvB
        OutputFlags _types;

        // vOCB
        String^ _pluginName;

        // t@CtB^B
        String^ _fileFilter;

        // vOCB
        String^ _pluginInfo;

        // ݒC^tF[XĂ邩B
        bool _hasConfig;

        // YUY2̃XP[O^CvB
        Yuy2ScaleType _yuy2Scale;

        // YUY2̐Fݒ^CvB
        Yuy2ColorMatrixType _yuy2ColorMatrix;

        // YUY2UV̗p@B
        Yuy2UVType _yuy2UV;

        // ݂ IOutputInfo (nullptr ȂΏĂȂ)B
        IOutputInfo^ _nowInfo;

        // ݂ OUTPUT_INFO B
        OUTPUT_INFO* _nowInfoValue;

        // ݂YUY2XP[O^CvB
        Yuy2ScaleType _nowYuy2Scale;
        // ݂YUY2̐Fݒ^CvB
        Yuy2ColorMatrixType _nowYuy2ColorMatrix;
        // ݂YUY2̗p@B
        Yuy2UVType _nowYuy2UV;

        // ݂̉摜f[^B
        unsigned char* _nowImage;
        // _nowImage ̃TCYB
        size_t _nowImageSize;

        // ݂̉f[^B
        unsigned char* _nowAudio;
        // _nowAudio ̃TCYB
        size_t _nowAudioSize;

        // o͒tOB
        bool _outputing;

        // o͊tOB
        bool _outputed;

        // A{[gptOB
        bool _forceAbort;

        // R[obNnhQ
        GCHandle _handleGetVideo;
        GCHandle _handleGetAudio;
        GCHandle _handleIsAbort;
        GCHandle _handleRestTimeDisp;
        GCHandle _handleGetFlag;
        GCHandle _handleUpdatePreview;
        GCHandle _handleGetVideoEx;
        bool _callbackAlloced;

    public:
        // ÓIRXgN^B
        static OutputPlugin()
        {
            // ݒt@CۑfBNgpX
            ConfigDataDirPath = Path::Combine(
                Environment::GetFolderPath(
                    Environment::SpecialFolder::LocalApplicationData),
                "ruche-home\\" + OutputPlugin::typeid->Namespace);
        }

        // RXgN^B
        // auoPath : o̓vOC̃t@CpXB
        OutputPlugin(String^ auoPath)
            :
            _filePath(nullptr),
            _fileName(nullptr),
            _hPlugin(NULL),
            _pi(NULL),
            _piRefCount(0),
            _types(),
            _pluginName(nullptr),
            _fileFilter(nullptr),
            _pluginInfo(nullptr),
            _hasConfig(false),
            _yuy2Scale(Yuy2ScaleType::Compressed),
            _yuy2ColorMatrix(Yuy2ColorMatrixType::Auto),
            _yuy2UV(Yuy2UVType::LeftOrigin),
            _nowInfo(nullptr),
            _nowInfoValue(NULL),
            _nowYuy2Scale(Yuy2ScaleType::Compressed),
            _nowYuy2ColorMatrix(Yuy2ColorMatrixType::Auto),
            _nowYuy2UV(Yuy2UVType::LeftOrigin),
            _nowImage(NULL),
            _nowImageSize(0),
            _nowAudio(NULL),
            _nowAudioSize(0),
            _outputing(false),
            _outputed(false),
            _forceAbort(false),
            _handleGetVideo(),
            _handleGetAudio(),
            _handleIsAbort(),
            _handleRestTimeDisp(),
            _handleGetFlag(),
            _handleUpdatePreview(),
            _handleGetVideoEx(),
            _callbackAlloced(false)
        {
            if (auoPath == nullptr)
            {
                throw gcnew ArgumentNullException("auoPath");
            }

            // tpXt@Cݒ
            _filePath = Path::GetFullPath(auoPath);
            _fileName = Path::GetFileName(_filePath);

            try
            {
                BeginPlugin();

                // o̓^Cv擾
                _types = static_cast<OutputFlags>(_pi->flag);

                // vOCݒ
                _pluginName = gcnew String(_pi->name);
                _pluginInfo = gcnew String(_pi->information);

                // tB^ݒ(\0 ؂肩 | ؂ɕύX)
                std::string filter;
                for (const char* s = _pi->filefilter; *s != '\0'; )
                {
                    filter += s;
                    s += ::strlen(s) + 1;
                    if (*s != '\0') { filter += "|"; }
                }
                _fileFilter = gcnew String(filter.data());

                // ݒC^tF[XĂ邩
                _hasConfig = (_pi->func_config != NULL);
            }
            finally
            {
                EndPlugin();
            }
        }

        // t@CiCUB
        !OutputPlugin()
        {
            // ̏I
            if (_nowInfo != nullptr)
            {
                EndOutput(_nowInfo);
            }

            // vOC
            while (_piRefCount > 0)
            {
                EndPlugin();
            }
        }

        // fXgN^(Dispose)B
        ~OutputPlugin()
        {
            this->!OutputPlugin();
        }

    private:
        // vOCC^tF[XLɂB
        // _hPlugin  _pi LɂȂB
        // G[̎͂Ȃ̂ŌĂяoŕK EndPlugin ĂԁB
        void BeginPlugin()
        {
            // Ăяoς݂ȂQƃJE^グ邾
            if (++_piRefCount >= 2) { return; }

            // std::wstring ɕϊ
            pin_ptr<wchar_t> pathPin = &_filePath->ToCharArray()[0];
            std::wstring path(pathPin, _filePath->Length);
            pathPin = nullptr;

            // vOCe[u
            PluginDataMap::iterator itr = PluginTable.find(path);
            if (itr != PluginTable.end())
            {
                // o^ς݂Ȃ̂łg

                _hPlugin = (*itr).second.getDllHandle();
                _pi = (*itr).second.getTable();
            }
            else
            {
                // ܂o^ĂȂ̂ŏ

                // W[擾
                _hPlugin = ::LoadLibraryW(path.data());
                if (_hPlugin == NULL)
                {
                    throw gcnew ApplicationException(
                        _fileName + " ͗LAviUtlo̓vOCł͂܂B");
                }

                // API֐擾
                Func_GetOutputPluginTable getOutputPluginTable =
                    reinterpret_cast<Func_GetOutputPluginTable>(
                        ::GetProcAddress(_hPlugin, "GetOutputPluginTable"));
                if (getOutputPluginTable == NULL)
                {
                    throw gcnew ApplicationException(
                        _fileName + " ͗LAviUtlo̓vOCł͂܂B");
                }

                // vOCC^tF[X擾
                _pi = getOutputPluginTable();
                if (_pi == NULL || _pi->func_output == NULL)
                {
                    throw gcnew ApplicationException(
                        _fileName + " C^tF[X擾ł܂łB");
                }

                // vOCC^tF[X
                if (_pi->func_init != NULL && _pi->func_init() == FALSE)
                {
                    throw gcnew ApplicationException(
                        _fileName + " ̏Ɏs܂B");
                }

                // vOCe[uɓo^
                PluginTable.insert(
                    PluginDataMap::value_type(path, PluginData(_hPlugin, _pi)));
            }
        }

        // vOCC^tF[X𖳌ɂB
        void EndPlugin()
        {
            // ܂QƂĂȂQƃJE^邾
            if (--_piRefCount >= 1) { return; }
            _piRefCount = 0;

            // vOCe[uĝŉ͂Ȃ
#if 0
            // vOCC^tF[X
            if (_pi != NULL)
            {
                if (_pi->func_exit != NULL)
                {
                    _pi->func_exit();
                }
                _pi = NULL;
            }

            // W[
            if (_hPlugin != NULL)
            {
                ::FreeLibrary(_hPlugin);
                _hPlugin = NULL;
            }
#else
            _pi = NULL;
            _hPlugin = NULL;
#endif
        }

    private:
        void* GetVideoProcess(int frame)
        {
            return GetVideoExProcess(frame, BI_RGB);
        }

        void* GetAudioProcess(int start, int length, int* readed)
        {
            // WindowsCxg
            Application::DoEvents();

            // f[^擾
            array<Byte>^ audio = _nowInfo->GetOutputAudio(start, length);

            // obt@
            size_t size = static_cast<size_t>(audio->Length);
            InitNowAudio(size);

            // Rs[
            pin_ptr<Byte> data = &audio[0];
            ::memcpy(_nowAudio, data, size);
            data = nullptr;

            // Tvݒ
            if (readed != NULL)
            {
                *readed = audio->Length / _nowInfo->OutputAudioSampleSize;
            }

            return _nowAudio;
        }

        BOOL IsAbortProcess()
        {
            return ((_forceAbort || _nowInfo->IsOutputAborted) ? TRUE : FALSE);
        }

        BOOL RestTimeDispProcess(int now, int total)
        {
            bool result = true;
            if (_nowInfo->FrameProgress != nullptr)
            {
                result = _nowInfo->FrameProgress(now, total);
            }
            return (result ? TRUE : FALSE);
        }

        int GetFlagProcess(int frame)
        {
            return (int)_nowInfo->GetVideoFrameFlags(frame);
        }

        BOOL UpdatePreviewProcess()
        {
            bool result = true;
            if (_nowInfo->UpdatePreview != nullptr)
            {
                result = _nowInfo->UpdatePreview();
            }
            return (result ? TRUE : FALSE);
        }

        void* GetVideoExProcess(int frame, DWORD format)
        {
            // WindowsCxg
            Application::DoEvents();

            // f[^擾(24rbgBGRgbv_E`)
            array<Byte>^ image = _nowInfo->GetOutputImage(frame);

            // _nowImage ֏o
            pin_ptr<Byte> data = &image[0];
            switch (format)
            {
            case mmioFOURCC('Y', 'U', 'Y', '2'):
                // YUY2
                PutYuy2Data(data);
                break;

            case mmioFOURCC('Y', 'C', '4', '8'):
                // PIXEL_YC
                PutPixelYcData(data);
                break;

            default:
                // BGR{gAbv
                PutBgrData(data);
                break;
            }
            data = nullptr;

            return _nowImage;
        }

        // _nowImage BGR{gAbv`ŏoB
        void PutBgrData(const unsigned char* src)
        {
            int w = _nowInfo->OutputImageWidth;
            int h = _nowInfo->OutputImageHeight;

            // obt@
            size_t size = static_cast<size_t>(_nowInfo->OutputImageFrameSize);
            InitNowImage(size);

            // sPʂŃRs[
            int lineLen = (w * 3 + 3) / 4 * 4;
            for (int sy = 0, dy = h - 1; sy < h; ++sy, --dy)
            {
                ::memcpy(
                    &_nowImage[dy * lineLen], &src[sy * lineLen], lineLen);
            }
        }

        // _nowImage YUY2`(AC2)ŏoB
        void PutYuy2Data(const unsigned char* src)
        {
            int r[2], g[2], b[2], rr, gg, bb, dy[2], du, dv;
            int yy[2], uu, vv, rate;

            int w = _nowInfo->OutputImageWidth;
            int h = _nowInfo->OutputImageHeight;
            Yuy2ColorMatrixType cmType =
                DecideYuy2ColorMatrixType(_nowYuy2ColorMatrix, h);

            // obt@
            size_t destLineLen = ((w + 1) / 2) * 4;
            size_t size = static_cast<size_t>(destLineLen * h);
            InitNowImage(size);

            // sPʂŏ
            int lineLen = (w * 3 + 3) / 4 * 4;
            for (int y = 0; y < h; ++y)
            {
                const unsigned char* ps = &src[y * lineLen];
                unsigned char* pd = &_nowImage[y * destLineLen];

                // 2sNZ
                for (int x = 0; x < w; x += 2)
                {
                    // 1sNZڎ擾
                    b[0] = ps[0];
                    g[0] = ps[1];
                    r[0] = ps[2];
                    ps += 3;

                    // 2sNZڎ擾
                    if (x < w - 1)
                    {
                        b[1] = ps[0];
                        g[1] = ps[1];
                        r[1] = ps[2];
                        ps += 3;
                    }
                    else
                    {
                        // ̃sNZȂŏIsNZ̒lp
                        b[1] = b[0];
                        g[1] = g[0];
                        r[1] = r[0];
                    }

                    // UVZopRGBI(l2{)
                    switch (_nowYuy2UV)
                    {
                    case Yuy2UVType::Average:
                        rr = r[0] + r[1];
                        gg = g[0] + g[1];
                        bb = b[0] + b[1];
                        break;

                    case Yuy2UVType::RightOrigin:
                        rr = r[1] * 2;
                        gg = g[1] * 2;
                        bb = b[1] * 2;
                        break;

                    case Yuy2UVType::LeftOrigin:
                    default:
                        rr = r[0] * 2;
                        gg = g[0] * 2;
                        bb = b[0] * 2;
                        break;
                    }

                    // FʂɃtWłYUVlƂ̔{p
                    // Ql: http://koujinz.cocolog-nifty.com/blog/2009/03/rgbycbcr-a4a5.html
                    switch (cmType)
                    {
                    case Yuy2ColorMatrixType::BT709:
                        for (int i = 0; i < 2; ++i)
                        {
                            yy[i] = (2126*r[i] + 7152*g[i] + 722*b[i]) * 2;
                        }
                        uu = -1146*rr - 3854*gg + 5000*bb + 255*10000;
                        vv = 5000*rr - 4542*gg - 458*bb + 255*10000;
                        rate = 2 * 10000;
                        break;

                    case Yuy2ColorMatrixType::BT601:
                    default:
                        for (int i = 0; i < 2; ++i)
                        {
                            yy[i] = (299*r[i] + 587*g[i] + 114*b[i]) * 2;
                        }
                        uu = -169*rr - 331*gg + 500*bb + 255*1000;
                        vv = 500*rr - 419*gg - 81*bb + 255*1000;
                        rate = 2 * 1000;
                        break;
                    }

                    // XP[OʂYUVlvZ
                    switch (_nowYuy2Scale)
                    {
                    case Yuy2ScaleType::Full:
                        for (int i = 0; i < 2; ++i)
                        {
                            dy[i] = static_cast<int>((yy[i] + (rate/2)) / rate);
                            dy[i] = min(max(0, dy[i]), 255);
                        }
                        du = static_cast<int>((uu + (rate/2)) / rate);
                        du = min(max(0, du), 255);
                        dv = static_cast<int>((vv + (rate/2)) / rate);
                        dv = min(max(0, dv), 255);
                        break;

                    case Yuy2ScaleType::Compressed:
                    default:
                        rate *= 255;
                        for (int i = 0; i < 2; ++i)
                        {
                            dy[i] = static_cast<int>(
                                (yy[i]*219 + (rate/2)) / rate + 16);
                            dy[i] = min(max(16, dy[i]), 235);
                        }
                        du = static_cast<int>((uu*224 + (rate/2)) / rate + 16);
                        du = min(max(16, du), 240);
                        dv = static_cast<int>((vv*224 + (rate/2)) / rate + 16);
                        dv = min(max(16, dv), 240);
                        break;
                    }

                    // lݒ
                    pd[0] = static_cast<unsigned char>(dy[0]);
                    pd[1] = static_cast<unsigned char>(du);
                    pd[2] = static_cast<unsigned char>(dy[1]);
                    pd[3] = static_cast<unsigned char>(dv);
                    pd += 4;
                }
            }
        }

        // _nowImage PIXEL_YC`ŏoB
        void PutPixelYcData(const unsigned char* src)
        {
            int r, g, b, dy, du, dv;

            int w = _nowInfo->OutputImageWidth;
            int h = _nowInfo->OutputImageHeight;

            // obt@
            size_t destLineLen = w * 6;
            size_t size = static_cast<size_t>(destLineLen * h);
            InitNowImage(size);

            // sPʂŏ
            int lineLen = (w * 3 + 3) / 4 * 4;
            for (int y = 0; y < h; ++y)
            {
                const unsigned char* ps = &src[y * lineLen];
                short* pd =
                    reinterpret_cast<short*>(&_nowImage[y * destLineLen]);

                // 1sNZ
                for (int x = 0; x < w; ++x)
                {
                    // sNZ擾
                    b = ps[0];
                    g = ps[1];
                    r = ps[2];
                    ps += 3;

                    // p[^vZ
                    // Ql: http://vm104.xen.klab.org/doc/aviutlyc.html
                    dy = (( 4918*r + 354) >> 10) + (( 9655*g + 585) >> 10) + (( 1875*b + 523) >> 10);
                    du = ((-2775*r + 240) >> 10) + ((-5449*g + 515) >> 10) + (( 8224*b + 256) >> 10);
                    dv = (( 8224*r + 256) >> 10) + ((-6887*g + 110) >> 10) + ((-1337*b + 646) >> 10);

                    // lݒ
                    pd[0] = static_cast<short>(dy);
                    pd[1] = static_cast<short>(du);
                    pd[2] = static_cast<short>(dv);
                    pd += 3;
                }
            }
        }

        // _nowImage B
        // size : VTCYB
        void InitNowImage(size_t size)
        {
            if (_nowImage == NULL || _nowImageSize < size)
            {
                if (_nowImage != NULL) { delete [] _nowImage; }
                _nowImage = new unsigned char[size];
                _nowImageSize = size;
            }
        }

        // _nowAudio B
        // size : VTCYB
        void InitNowAudio(size_t size)
        {
            if (_nowAudio == NULL || _nowAudioSize < size)
            {
                if (_nowAudio != NULL) { delete [] _nowAudio; }
                _nowAudio = new unsigned char[size];
                _nowAudioSize = size;
            }
        }

    private:
        // R[obNB
        void AllocOutputCallbacks()
        {
            if (!_callbackAlloced)
            {
                // fQ[g쐬
                GetVideoCallback^ cbGetVideo =
                    gcnew GetVideoCallback(
                        this, &OutputPlugin::GetVideoProcess);
                GetAudioCallback^ cbGetAudio =
                    gcnew GetAudioCallback(
                        this, &OutputPlugin::GetAudioProcess);
                IsAbortCallback^ cbIsAbort =
                    gcnew IsAbortCallback(
                        this, &OutputPlugin::IsAbortProcess);
                RestTimeDispCallback^ cbRestTimeDisp =
                    gcnew RestTimeDispCallback(
                        this, &OutputPlugin::RestTimeDispProcess);
                GetFlagCallback^ cbGetFlag =
                    gcnew GetFlagCallback(
                        this, &OutputPlugin::GetFlagProcess);
                UpdatePreviewCallback^ cbUpdatePreview =
                    gcnew UpdatePreviewCallback(
                        this, &OutputPlugin::UpdatePreviewProcess);
                GetVideoExCallback^ cbGetVideoEx =
                    gcnew GetVideoExCallback(
                        this, &OutputPlugin::GetVideoExProcess);

                // nh쐬
                _handleGetVideo      = GCHandle::Alloc(cbGetVideo);
                _handleGetAudio      = GCHandle::Alloc(cbGetAudio);
                _handleIsAbort       = GCHandle::Alloc(cbIsAbort);
                _handleRestTimeDisp  = GCHandle::Alloc(cbRestTimeDisp);
                _handleGetFlag       = GCHandle::Alloc(cbGetFlag);
                _handleUpdatePreview = GCHandle::Alloc(cbUpdatePreview);
                _handleGetVideoEx    = GCHandle::Alloc(cbGetVideoEx);

                _callbackAlloced = true;
            }
        }

        // R[obNB
        void FreeOutputCallbacks()
        {
            if (_callbackAlloced)
            {
                // nhj
                _handleGetVideo.Free();
                _handleGetAudio.Free();
                _handleIsAbort.Free();
                _handleRestTimeDisp.Free();
                _handleGetFlag.Free();
                _handleUpdatePreview.Free();
                _handleGetVideoEx.Free();

                _callbackAlloced = false;
            }
        }

        // R[obN OUTPUT_INFO ɐݒ肷B
        void SetupOutputCallbacks(OUTPUT_INFO* oi)
        {
            AllocOutputCallbacks();

            // IntPtr擾
            IntPtr ptrGetVideo =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleGetVideo.Target);
            IntPtr ptrGetAudio =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleGetAudio.Target);
            IntPtr ptrIsAbort =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleIsAbort.Target);
            IntPtr ptrRestTimeDisp =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleRestTimeDisp.Target);
            IntPtr ptrGetFlag =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleGetFlag.Target);
            IntPtr ptrUpdatePreview =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleUpdatePreview.Target);
            IntPtr ptrGetVideoEx =
                Marshal::GetFunctionPointerForDelegate(
                    (System::Delegate^)_handleGetVideoEx.Target);

            // R[obNݒ
            oi->func_get_video =
                static_cast<void* (*)(int)>(ptrGetVideo.ToPointer());
            oi->func_get_audio =
                static_cast<void* (*)(int, int, int*)>(ptrGetAudio.ToPointer());
            oi->func_is_abort =
                static_cast<BOOL (*)()>(ptrIsAbort.ToPointer());
            oi->func_rest_time_disp =
                static_cast<BOOL (*)(int, int)>(ptrRestTimeDisp.ToPointer());
            oi->func_get_flag =
                static_cast<int (*)(int)>(ptrGetFlag.ToPointer());
            oi->func_update_preview =
                static_cast<BOOL (*)()>(ptrUpdatePreview.ToPointer());
            oi->func_get_video_ex =
                static_cast<void* (*)(int, DWORD)>(ptrGetVideoEx.ToPointer());
        }

        // o͏̊JnʒmA OUTPUT_INFO 𓾂B
        // info : ƂȂ IOutputInfo B
        // yuy2Scale : YUY2ϊsꍇ̃XP[O^CvB
        // yuy2ColorMatrix : YUY2ϊsꍇ̐Fݒ^CvB
        // yuy2UV : YUY2ϊsꍇUV̗p@B
        // return:
        //      OUTPUT_INFO |C^B
        //      ݑ̏o͏sĂꍇ NULL B
        OUTPUT_INFO* BeginOutput(
            IOutputInfo^ info,
            Yuy2ScaleType yuy2Scale,
            Yuy2ColorMatrixType yuy2ColorMatrix,
            Yuy2UVType yuy2UV)
        {
            // ̏o͏sĂB
            if (_nowInfo != nullptr) { return NULL; }

            // _nowInfo ݒ
            if (info == nullptr)
            {
                throw gcnew ArgumentNullException("info");
            }
            _nowInfo = info;

            // OUTPUT_INFO \z
            OUTPUT_INFO* oi = _nowInfoValue = new OUTPUT_INFO;
            ::ZeroMemory(oi, sizeof(OUTPUT_INFO));

            // OUTPUT_INFO oϐݒ
            oi->flag        = static_cast<int>(info->OutputTypes);
            oi->w           = info->OutputImageWidth;
            oi->h           = info->OutputImageHeight;
            oi->rate        = info->OutputImageFrameRateBase;
            oi->scale       = info->OutputImageFrameRateScale;
            oi->n           = info->OutputImageFrameCount;
            oi->size        = info->OutputImageFrameSize;
            oi->audio_rate  = info->OutputAudioSampleRate;
            oi->audio_ch    = info->OutputAudioChannel;
            oi->audio_n     = info->OutputAudioSampleCount;
            oi->audio_size  = info->OutputAudioSampleSize;
            IntPtr fnPtr    = Marshal::StringToHGlobalAnsi(info->OutputFilePath);
            oi->savefile    = static_cast<char*>(fnPtr.ToPointer());

            // OUTPUT_INFO R[obNݒ
            SetupOutputCallbacks(oi);

            // YUY2֘Aݒ
            _nowYuy2Scale = yuy2Scale;
            _nowYuy2ColorMatrix = yuy2ColorMatrix;
            _nowYuy2UV = yuy2UV;

            // tO
            _outputing = false;
            _outputed = false;
            _forceAbort = false;

            return oi;
        }

        // o͏̏Iʒm܂B
        bool EndOutput(IOutputInfo^ info)
        {
            bool result = false;

            if (_nowInfo != nullptr && _nowInfo == info)
            {
                // o͒Ȃ狭I
                _forceAbort = true;
                if (!_outputed && _outputing)
                {
                    while (!_outputed) { Thread::Sleep(1); }
                }

                // obt@
                if (_nowImage != NULL)
                {
                    delete [] _nowImage;
                    _nowImage = NULL;
                }
                _nowImageSize = 0;
                if (_nowAudio != NULL)
                {
                    delete [] _nowAudio;
                    _nowAudio = NULL;
                }
                _nowAudioSize = 0;

                // OUTPUT_INFO R[obN
                FreeOutputCallbacks();

                // OUTPUT_INFO 
                OUTPUT_INFO* oi = _nowInfoValue;
                if (oi != NULL)
                {
                    if (oi->savefile != NULL)
                    {
                        Marshal::FreeHGlobal(IntPtr(oi->savefile));
                    }

                    delete _nowInfoValue;
                    _nowInfoValue = NULL;
                }

                _nowInfo = nullptr;
            }

            return result;
        }

    public:
        // vOC̃t@CpX擾B
        property String^ PluginFilePath
        {
            String^ get() { return _filePath; }
        }

        // vOC̃t@C擾B
        property String^ PluginFileName
        {
            String^ get() { return _fileName; }
        }

        // vOC̐ݒۑpt@C̃pX擾B
        property String^ PluginConfigFilePath
        {
            String^ get()
            {
                return Path::Combine(
                    ConfigDataDirPath, PluginFileName + ".auoconf");
            }
        }

        // o̓^CvtO擾B
        property OutputFlags OutputTypes
        {
            OutputFlags get() { return _types; }
        }

        // vOC擾B
        property String^ PluginName
        {
            String^ get() { return _pluginName; }
        }

        // t@CtB^擾B
        property String^ FileFilter
        {
            String^ get() { return _fileFilter; }
        }

        // vOC擾B
        property String^ PluginInfo
        {
            String^ get() { return _pluginInfo; }
        }

        // ݒC^tF[XĂ邩ۂ擾B
        property bool HasConfig
        {
            bool get() { return _hasConfig; }
        }

        // YUY2ϊsꍇ̃XP[O^Cv擾܂͐ݒ肷B
        property Yuy2ScaleType Yuy2Scale
        {
            Yuy2ScaleType get() { return _yuy2Scale; }
            void set(Yuy2ScaleType value) { _yuy2Scale = value; }
        };

        // YUY2ϊsꍇ̐Fݒ^Cv擾܂͐ݒ肷B
        property Yuy2ColorMatrixType Yuy2ColorMatrix
        {
            Yuy2ColorMatrixType get() { return _yuy2ColorMatrix; }
            void set(Yuy2ColorMatrixType value) { _yuy2ColorMatrix = value; }
        };

        // YUY2ϊsꍇUV̗̍p@擾܂͐ݒ肷B
        property Yuy2UVType Yuy2UV
        {
            Yuy2UVType get() { return _yuy2UV; }
            void set(Yuy2UVType value) { _yuy2UV = value; }
        };

        // ݒt@Cݒf[^ǂݎAvOCɐݒ肷B
        bool LoadPluginConfig()
        {
            bool result = false;

            if (File::Exists(PluginConfigFilePath))
            {
                try
                {
                    BeginPlugin();

                    if (_pi->func_config_set != NULL)
                    {
                        // t@Cǂݎ
                        array<Byte>^ data =
                            File::ReadAllBytes(PluginConfigFilePath);

                        // vOC֐ݒ
                        pin_ptr<unsigned char> dataPtr = &data[0];
                        int useSize =
                            _pi->func_config_set(dataPtr, data->Length);
                        result = (useSize > 0);
                        dataPtr = nullptr;
                    }
                }
                catch (...)
                {
                    result = false;
                }
                finally
                {
                    EndPlugin();
                }
            }

            return result;
        }

        // vOCݒ擾Aݒt@CɏoB
        bool SavePluginConfig()
        {
            bool result = false;

            try
            {
                BeginPlugin();

                if (_pi->func_config_get != NULL)
                {
                    // KvTCY𓾂
                    int needSize = _pi->func_config_get(NULL, 0);
                    if (needSize > 0)
                    {
                        // vOCf[^ǂݎ
                        array<Byte>^ data = gcnew array<Byte>(needSize);
                        pin_ptr<unsigned char> dataPtr = &data[0];
                        _pi->func_config_get(dataPtr, needSize);
                        dataPtr = nullptr;

                        String^ dirPath =
                            Path::GetDirectoryName(PluginConfigFilePath);
                        if (dirPath != nullptr)
                        {
                            // t@CۑfBNgȂ΍쐬
                            if (!Directory::Exists(dirPath))
                            {
                                Directory::CreateDirectory(dirPath);
                            }

                            // t@C֕ۑ
                            File::WriteAllBytes(PluginConfigFilePath, data);

                            result = true;
                        }
                    }
                }
            }
            catch (...)
            {
                result = false;
            }
            finally
            {
                EndPlugin();
            }

            return result;
        }

        // ݒC^tF[X\B
        // hwnd : eEBhẼnhB nullptr ł悢B
        bool ShowConfig(IntPtr^ handle)
        {
            bool result = false;

            if (HasConfig)
            {
                // nh擾
                HWND hwnd = NULL;
                if (handle != nullptr)
                {
                    hwnd = static_cast<HWND>(handle->ToPointer());
                }

                try
                {
                    BeginPlugin();

                    // ݒǂݎ
                    LoadPluginConfig();

                    // Ăяo
                    result = (_pi->func_config(hwnd, _hPlugin) != FALSE);
                    if (result)
                    {
                        // ݒۑ
                        SavePluginConfig();
                    }
                }
                catch (...)
                {
                    result = false;
                    throw;
                }
                finally
                {
                    EndPlugin();
                }
            }

            return result;
        }

        // o͏ɁAf[^o͂B
        // info : o͏B
        bool Execute(IOutputInfo^ info)
        {
            bool result = false;

            // OUTPUT_INFO 쐬
            OUTPUT_INFO* oi =
                BeginOutput(info, Yuy2Scale, Yuy2ColorMatrix, Yuy2UV);
            if (oi != NULL)
            {
                try
                {
                    BeginPlugin();

                    // ݒǂݎ
                    LoadPluginConfig();

                    // o͏s
                    _outputing = true;
                    result = (_pi->func_output(oi) != FALSE);
                }
                finally
                {
                    // o͊
                    _outputed = true;

                    // I
                    if (!_forceAbort)
                    {
                        EndOutput(info);
                    }

                    EndPlugin();
                }
            }

            return result;
        }

        // ̃vOC̕\擾B
        virtual String^ ToString() override
        {
            return PluginName;
        }
    };

    ////////////////////
}
}}
