using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using nicoAE.Plugin.Interface;
using nicoAE.Plugin.Utils;
using ruche.aviutl.plugins;

namespace ruche.nive.aviutlPluginPorts
{
    /// <summary>
    /// AviUtl̓vOC𗘗pNiVErfI̓vOCNXB
    /// MediaInput ō쐬B
    /// </summary>
    internal class VideoMediaInput : IMediaInput, IInputChild
    {
        /// <summary>
        /// eCX^XB
        /// </summary>
        private IMediaInput _parent;

        /// <summary>
        /// ̓vOCC^tF[XB
        /// </summary>
        private InputPlugin _plugin;

        /// <summary>
        /// vWFNgB
        /// </summary>
        private IHost _host;

        /// <summary>
        /// ʔԍB
        /// </summary>
        private long _inputCode;

        /// <summary>
        /// RXgN^B
        /// </summary>
        /// <param name="parent">eCX^XB</param>
        /// <param name="plugin">
        /// fBAt@Cǂݍݍς݂̓̓vOCB
        /// </param>
        /// <param name="host">vWFNgB</param>
        /// <param name="id">q̎IDB</param>
        public VideoMediaInput(
            IMediaInput parent, InputPlugin plugin, IHost host, int id)
        {
            if (parent == null)
            {
                throw new NullReferenceException("parent  null łB");
            }
            if (plugin == null)
            {
                throw new NullReferenceException("plugin  null łB");
            }
            if (host == null)
            {
                throw new NullReferenceException("host  null łB");
            }
            if (!plugin.IsMediaFileOpened)
            {
                throw new ArgumentException(
                    "fBAt@CJĂ܂B", "plugin");
            }

            _parent = parent;
            _plugin = plugin;
            _host = host;
            _inputCode =
                ((long)_plugin.MediaFilePath.GetHashCode() << 32) + id;
        }

        /// <summary>
        /// ͂̊JnsB
        /// </summary>
        void BeginInput()
        {
            // vOĈfBNgֈړ
            CurrentDirectoryStack.Instance.Push(
                Path.GetDirectoryName(_plugin.PluginFilePath));
        }

        /// <summary>
        /// ͂̏IsB
        /// </summary>
        void EndInput()
        {
            // JgfBNg
            CurrentDirectoryStack.Instance.Pop();
        }

        #region IMediaInput o

        public string PluginName
        {
            get { return "AviUtl̓vOCp"; }
        }

        public string CorrespondingType
        {
            get { return "fBAt@C (*.*)|*.*"; }
        }

        public string FileName
        {
            get { return _plugin.MediaFilePath; }
        }

        public void Initialize(IHost host)
        {
            // gpȂ
        }

        public bool LoadFile(string file)
        {
            // gpȂ
            return false;
        }

        public MediaInfo GetMediaInfo()
        {
            IInputInfo info = _plugin.MediaFileInfo;

            return new MediaInfo(
                info.InputImageFrameCount,
                new Size(info.InputImageWidth, info.InputImageHeight),
                MediaType.Video,
                (double)info.InputImageFrameRateBase / info.InputImageFrameRateScale);
        }

        public int GetLength()
        {
            IInputInfo info = _plugin.MediaFileInfo;

            double ratio =
                (_host.ProjectFrameRate * info.InputImageFrameRateScale) /
                info.InputImageFrameRateBase;
            return (int)(info.InputImageFrameCount * ratio);
        }

        public Bitmap GetFrame(int index)
        {
            Bitmap bmp = null;

            BeginInput();

            // 擾
            IInputInfo info = _plugin.MediaFileInfo;
            int width = info.InputImageWidth;
            int height = info.InputImageHeight;

            try
            {
                // ǂݍނׂt[ʒuZo
                double baseFrameRate =
                    (double)info.InputImageFrameRateBase /
                    info.InputImageFrameRateScale;
                int frame = PluginUtils.FrameIndex(
                    baseFrameRate, _host.ProjectFrameRate, index);

                // ǂݎ
                byte[] buf = _plugin.ReadMediaFileImage(frame);
                if (buf != null)
                {
                    // 32rbgARGBrbg}bv쐬
                    bmp = new Bitmap(width, height, PixelFormat.Format32bppArgb);

                    // sNZf[^ݒ
                    BitmapData bmpData = bmp.LockBits(
                        new Rectangle(0, 0, width, height),
                        ImageLockMode.WriteOnly,
                        PixelFormat.Format32bppArgb);
                    Marshal.Copy(buf, 0, bmpData.Scan0, (width * 4) * height);
                    bmp.UnlockBits(bmpData);
                }
            }
            finally
            {
                EndInput();

                if (bmp == null)
                {
                    bmp = PluginUtils.NullImage(width, height);
                }
            }

            return bmp;
        }

        public byte[] GetAudio()
        {
            return null;
        }

        public byte[] GetFrameAudio(int index)
        {
            return null;
        }

        #endregion

        #region IDisposable o

        public void Dispose()
        {
            _plugin = null;
        }

        #endregion

        #region IInputChild o

        public long InputCode
        {
            get { return _inputCode; }
        }

        public IMediaInput Parent
        {
            get { return _parent; }
        }

        #endregion
    }
}
