using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using nicoAE.Plugin.Interface;
using nicoAE.Plugin.Utils;
using ruche.aviutl.plugins;

namespace ruche.nive.aviutlPluginPorts
{
    /// <summary>
    /// AviUtlo̓vOC𗘗pNiVEo̓vOCNXB
    /// </summary>
    public class MediaOutput : IMediaOutput, IOutputInfo, IHavePreviewImage, IAbortable
    {
        /// <summary>
        /// 摜t@C̃vtBNXB
        /// </summary>
        public const string ImageFileNamePrefix = "__image_";

        /// <summary>
        /// t@C̃vtBNXB
        /// </summary>
        public const string AudioFileNamePrefix = "__audio_";

        /// <summary>
        /// ꕔ\bhŗpC[W]pf[^B
        /// </summary>
        private static readonly byte[] TempImageBlankData = { 0, 0, 0, 0 };

        /// <summary>
        /// f[^1b̕WTCYB
        /// </summary>
        private static readonly int AudioSizePerSecond = (int)(
            ReferenceData.WaveFormat.nSamplesPerSec *
            ReferenceData.WaveFormat.nBlockAlign);

        /// <summary>
        /// ƃfBNg̃tpXB
        /// </summary>
        private string _workDirPath = null;

        /// <summary>
        /// ƃfBNgV쐬ꂽȂ true B
        /// </summary>
        private bool _workDirCreated = false;

        /// <summary>
        /// 摜t@C̃tpX̃vtBNXB
        /// </summary>
        private string _imageFilePathPrefix = null;

        /// <summary>
        /// t@C̃tpX̃vtBNXB
        /// </summary>
        private string _audioFilePathPrefix = null;

        /// <summary>
        /// 摜t@C̊gq(sIht)B
        /// </summary>
        private string _imageFileExtension = ".png";

        /// <summary>
        /// 摜t@C̏o̓tH[}bgB
        /// </summary>
        private WorkImageType _imageFileType = WorkImageType.Png;

        /// <summary>
        /// 摜t@Č`BƎ`Ȃ null B
        /// </summary>
        private ImageFormat _imageFormat = ImageFormat.Png;

        /// <summary>
        /// rfIC[W̃TCYB
        /// </summary>
        private Size _imageSize;

        /// <summary>
        /// t[[gB
        /// </summary>
        private Fraction _frameRate;

        /// <summary>
        /// t[B
        /// </summary>
        private int _frameCount;

        /// <summary>
        /// o̓vOC̃C^tF[XB
        /// </summary>
        private OutputPlugin _plugin;

        /// <summary>
        /// ŏIIɏo͂t@C̃pXB
        /// </summary>
        private string _outputFilePath = "";

        /// <summary>
        /// o͌ɒԃt@C폜邩ۂB
        /// </summary>
        private bool _deleteWork = true;

        /// <summary>
        /// o͍ς݂̉摜t[B
        /// </summary>
        private int _imageFrameCount = 0;

        /// <summary>
        /// o͍ς݂̉t[B
        /// </summary>
        private int _audioFrameCount = 0;

        /// <summary>
        /// ŌɎ擾ꂽC[WB
        /// </summary>
        private Bitmap _lastImage = null;

        /// <summary>
        /// ŌɎ擾ꂽC[Wf[^B
        /// </summary>
        private byte[] _lastImageData = null;

        /// <summary>
        /// ŌɎ擾ꂽI[fBIf[^B
        /// </summary>
        private byte[] _lastAudioData = null;

        /// <summary>
        /// o͏̃A{[g邩ۂB
        /// </summary>
        private bool _outputAborted = false;

        /// <summary>
        /// vOCo͍ς݂ۂB
        /// </summary>
        private bool _executed = false;

        /// <summary>
        /// t[iʒm̂߂̃fQ[gB
        /// </summary>
        private FrameProgressDelegate _frameProgress = null;

        /// <summary>
        /// vr[XVv̂߂̃fQ[gB
        /// </summary>
        private UpdatePreviewDelegate _updatePreview = null;

#if OUTPUT_WORK_TIME
        /// <summary>
        /// ԃf[^o͂̊Jn
        /// </summary>
        private DateTime _workStartTime = DateTime.Now;
#endif // OUTPUT_WORK_TIME

        /// <summary>
        /// ꕔ\bhŗpC[Wf[^obt@B
        /// </summary>
        private static byte[] _tempImageData = null;

        /// <summary>
        /// ꕔ\bhŗpC[Wf[^obt@̃bNIuWFNgB
        /// </summary>
        private static Object _tempImageDataLock = new Object();

        /// <summary>
        /// RXgN^B
        /// </summary>
        public MediaOutput()
        {
        }

        /// <summary>
        /// fXgN^B
        /// </summary>
        ~MediaOutput()
        {
            Dispose();
        }

        /// <summary>
        /// o̓vOCB
        /// </summary>
        private void FreeOutputPlugin()
        {
            if (_plugin != null)
            {
                try
                {
                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    _plugin.Dispose();
                    _plugin = null;
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        /// <summary>
        /// ƃfBNg݂Ȃ΍쐬B
        /// </summary>
        /// <returns>
        /// 쐬ꂽꍇ true B݂ꍇ false B
        /// </returns>
        private bool CreateWorkDirectory()
        {
            bool result = false;

            if (!Directory.Exists(_workDirPath))
            {
                Directory.CreateDirectory(_workDirPath);
                _workDirCreated = true;
                result = true;
            }

            return result;
        }

        /// <summary>
        /// w肵t[ۑ摜t@C̃pX擾܂B
        /// </summary>
        /// <param name="frame">t[B</param>
        /// <returns>摜t@C̃pXB</returns>
        private string GetImageFilePath(int frame)
        {
            return _imageFilePathPrefix + frame + _imageFileExtension;
        }

        /// <summary>
        /// w肵t[ۑ鉹t@C̃pX擾܂B
        /// </summary>
        /// <param name="frame">t[B</param>
        /// <returns>t@C̃pXB</returns>
        private string GetAudioFilePath(int frame)
        {
            return _audioFilePathPrefix + frame + ".sndpart";
        }

        /// <summary>
        /// ɏoׂ摜t@C̃pX擾܂B
        /// </summary>
        private string OutputImageFilePath
        {
            get { return GetImageFilePath(_imageFrameCount); }
        }

        /// <summary>
        /// ɏoׂt@C̃pX擾܂B
        /// </summary>
        private string OutputAudioFilePath
        {
            get { return GetAudioFilePath(_audioFrameCount); }
        }

        /// <summary>
        /// f[^1t[̘_TCY擾܂B
        /// </summary>
        private double AudioFrameLogicalSize
        {
            get
            {
                return (
                    (double)AudioSizePerSecond * _frameRate.Denominator /
                    _frameRate.Numerator);
            }
        }

        /// <summary>
        /// f[^1t[̎TCY擾܂B
        /// K 4 ̔{ɂȂ܂B
        /// </summary>
        private int AudioFrameSize
        {
            get
            {
                int size = (int)(
                    (long)AudioSizePerSecond * _frameRate.Denominator /
                    _frameRate.Numerator);
                return (size / 4 * 4);
            }
        }

        /// <summary>
        /// f[^1t[̎Tv擾܂B
        /// </summary>
        private int AudioSampleCountPerFrame
        {
            get { return (AudioFrameSize / 4); }
        }

        /// <summary>
        /// w肵oCgʒut[ڂɂ邩ZoB
        /// </summary>
        /// <param name="length">oCgʒuB</param>
        /// <returns>t[B</returns>
        private double CalcAudioFrameFromBytes(double length)
        {
            return (length / AudioFrameLogicalSize);
        }

        /// <summary>
        /// Is܂B
        /// </summary>
        public void Finish()
        {
            try
            {
                // vOC
                FreeOutputPlugin();

                // \[X
                if (_lastImage != null)
                {
                    Bitmap tmp = _lastImage;
                    _lastImage = null;
                    tmp.Dispose();
                }
                _lastImageData = null;
                _lastAudioData = null;

                if (_deleteWork)
                {
                    // t@C폜
                    while (--_imageFrameCount >= 0)
                    {
                        if (File.Exists(OutputImageFilePath))
                        {
                            File.Delete(OutputImageFilePath);
                        }
                    }
                    while (--_audioFrameCount >= 0)
                    {
                        if (File.Exists(OutputAudioFilePath))
                        {
                            File.Delete(OutputAudioFilePath);
                        }
                    }

                    // VK쐬fBNgȂ΍폜
                    if (_workDirCreated)
                    {
                        if (Directory.Exists(_workDirPath))
                        {
                            // łȂΎs
                            try
                            {
                                Directory.Delete(_workDirPath, false);
                            }
                            catch { }
                        }
                    }
                }

                // t[
                _imageFrameCount = 0;
                _audioFrameCount = 0;
            }
            catch { }
        }

        /// <summary>
        /// 32rbgBMPC[WBGR24Ǝ`ŃXg[ɏoB
        /// </summary>
        /// <param name="stream">õXg[B</param>
        /// <param name="image">o32rbgBMPC[WB</param>
        private static void WriteBgr24Image(Stream stream, Bitmap image)
        {
            lock (_tempImageDataLock)
            {
                int width = image.Width;
                int height = image.Height;

                // o1C̃oCgƎTCY
                int destLineLen = width * 3;
                int destLineSize = (destLineLen + 3) & ~0x03; // 4 bytes align
                int destLineBlank = destLineSize - destLineLen;

                // ꎞobt@
                int srcSize = width * height * 4;
                if (_tempImageData == null || _tempImageData.Length < srcSize)
                {
                    _tempImageData = new byte[srcSize];
                }

                // 32rbgBGRAf[^ꎞobt@ɃRs[
                BitmapData bmpData = image.LockBits(
                    new Rectangle(0, 0, width, height),
                    ImageLockMode.ReadOnly, PixelFormat.Format32bppArgb);
                Marshal.Copy(bmpData.Scan0, _tempImageData, 0, srcSize);
                image.UnlockBits(bmpData);

                // Xg[24rbgBGRƂďo
                for (int y = 0, srcPos = 0; y < height; ++y)
                {
                    // l3oCgo
                    for (int x = 0; x < width; ++x)
                    {
                        stream.Write(_tempImageData, srcPos, 3);
                        srcPos += 4;
                    }

                    // ACp]o
                    if (destLineBlank > 0)
                    {
                        stream.Write(TempImageBlankData, 0, destLineBlank);
                    }
                }
            }
        }

        #region IMediaOutput o

        public string PluginName
        {
            get { return "AviUtlo̓vOCp"; }
        }

        public nicoAE.Plugin.Utils.MediaType InputType
        {
            get { return MediaType.Video; }
        }

        public string CorrespondingType
        {
            get { return "ƃtH_͂Ă|*"; }
        }

        public string FileName
        {
            get { return _outputFilePath; }
        }

        public void Initialize(string workDirPath, IHost host)
        {
            // ƃfBNg̐ݒ(Ȃ΍쐬)
            for (int i = 0; ; ++i)
            {
                try
                {
                    _workDirPath =
                        (i == 0) ? workDirPath : (workDirPath + i);
                    CreateWorkDirectory();
                }
                catch (IOException ex)
                {
                    // IOException 100ȉȂpXςčăgC
                    if (i < 99) { continue; }
                    throw ex;
                }
                break;
            }

            // t@CpX̃vtBNXݒ
            _imageFilePathPrefix =
                Path.Combine(_workDirPath, ImageFileNamePrefix);
            _audioFilePathPrefix =
                Path.Combine(_workDirPath, AudioFileNamePrefix);

            // vWFNg̕ێ
            _imageSize = host.ProjectSize;
            _frameCount = host.ProjectLength;
            try
            {
                if (host.ProjectFrameRate <= 0.0)
                {
                    throw new ArgumentOutOfRangeException();
                }
                _frameRate = new Fraction(host.ProjectFrameRate);
            }
            catch (ArgumentOutOfRangeException)
            {
                // FPSeLg[ɐݒ
                _frameRate.Numerator =
                    (host.ProjectFrameRate <= 1.0) ? 1 : 960;
                _frameRate.Denominator = 1;
            }

            // Cuݒ̃[h
            LibraryConfig.Datas.Load();

            _executed = false;
        }

        public bool ShowSaveOption(IntPtr hwnd)
        {
            bool ok = false;

            try
            {
                using (MediaOutputConfigDialog dlg = new MediaOutputConfigDialog())
                {
                    // ݒ_CAO\
                    ok = (dlg.ShowDialog() == DialogResult.OK);
                    if (ok)
                    {
                        // 擾
                        FreeOutputPlugin();
                        _plugin = dlg.Plugin;
                        _outputFilePath = dlg.FileName;
                        _deleteWork = dlg.DeleteWork;
                        _imageFileType = dlg.WorkImageType;
                        _imageFileExtension =
                            "." + WorkImageTypeUtil.GetExtension(_imageFileType);
                        _imageFormat =
                            WorkImageTypeUtil.GetImageFormat(_imageFileType);
                    }
                }
            }
            catch (ArgumentException)
            {
                ok = false;
            }

            if (!ok) { Finish(); }

#if OUTPUT_WORK_TIME
            // ԌvJn
            if (ok)
            {
                _workStartTime = DateTime.Now;
            }
#endif // OUTPUT_WORK_TIME

            return ok;
        }

        public void AddFrameImage(Bitmap image)
        {
            // ꉞAƃfBNg쐬
            CreateWorkDirectory();

            // 摜o
            using (FileStream fs = new FileStream(
                OutputImageFilePath, FileMode.Create, FileAccess.Write))
            {
                if (_imageFormat != null)
                {
                    // W`
                    image.Save(fs, _imageFormat);
                }
                else
                {
                    // Ǝ`
                    switch (_imageFileType)
                    {
                        case WorkImageType.Bgr24:
                            WriteBgr24Image(fs, image);
                            break;

                        default:
                            // ɂ͂Ȃ͂c
                            throw new InvalidOperationException(
                                "ԉ摜t@C`słB");
                    }
                }
                fs.Close();
            }

            ++_imageFrameCount;
        }

        public void AddFrameAudio(byte[] audio)
        {
            // ꉞAƃfBNg쐬
            CreateWorkDirectory();

            // o
            using (FileStream fs = new FileStream(
                OutputAudioFilePath, FileMode.Create, FileAccess.Write))
            {
                fs.Write(audio, 0, audio.Length);
                fs.Close();
            }

            ++_audioFrameCount;
        }

        public void Close()
        {
            // t[oƂɂȂ̂mpȂ̂ŁA
            // Close Ă΂ꂽ_ŏoĂf[^
            // AviUtlo̓vOCɓɂȂĂ܂B
            try
            {
                _outputAborted = false;

                if (
                    !_executed &&
                    (OutputImageFrameCount  > 0 ||
                     OutputAudioSampleCount > 0))
                {
#if OUTPUT_WORK_TIME
                    // vvԂt@Cɏo
                    TimeSpan span = DateTime.Now - _workStartTime;
                    File.AppendAllText(
                        OutputFilePath + "_worktime.txt",
                        "  WORK TIME : " + span.ToString() + "\n",
                        Encoding.Default);
#endif // OUTPUT_WORK_TIME

                    using (AviutlOutputDialog dlg =
                        new AviutlOutputDialog(_plugin, this, this, this))
                    {
                        dlg.ShowDialog();
                    }
                }
            }
            catch (Exception ex)
            {
            #if DEBUG
                MessageBox.Show(
                    ex.Message + "\n\n" + ex.StackTrace, ex.GetType().FullName,
                    MessageBoxButtons.OK);
            #else
                MessageBox.Show(ex.Message, "G[", MessageBoxButtons.OK);
            #endif
            }
            finally
            {
                // 
                _executed = true;
                _outputAborted = false;

                // I
                Finish();
            }
        }

        #endregion

        #region IDisposable o

        public void Dispose()
        {
            Finish();
        }

        #endregion

        #region IOutputInfo o

        public OutputFlags OutputTypes
        {
            get
            {
                OutputFlags types = OutputFlags.Video;
                if (_audioFrameCount > 0)
                {
                    types |= OutputFlags.Audio;
                }
                return types;
            }
        }

        public int OutputImageWidth
        {
            get { return _imageSize.Width; }
        }

        public int OutputImageHeight
        {
            get { return _imageSize.Height; }
        }

        public int OutputImageFrameRateBase
        {
            get { return _frameRate.Numerator; }
        }

        public int OutputImageFrameRateScale
        {
            get { return _frameRate.Denominator; }
        }

        public int OutputImageFrameCount
        {
            get { return _imageFrameCount; }
        }

        public int OutputImageFrameSize
        {
            get
            {
                return (
                    ((_imageSize.Width * 3 + 3) / 4 * 4) * _imageSize.Height);
            }
        }

        public int OutputAudioSampleRate
        {
            get
            {
                return (int)(
                    (double)AudioSampleCountPerFrame * _frameRate.Numerator /
                    _frameRate.Denominator);
            }
        }

        public int OutputAudioChannel
        {
            get { return 2; }
        }

        public int OutputAudioSampleCount
        {
            get { return (AudioSampleCountPerFrame * _audioFrameCount); }
        }

        public int OutputAudioSampleSize
        {
            get { return 4; }
        }

        public string OutputFilePath
        {
            get { return FileName; }
        }

        public bool IsOutputAborted
        {
            get { return _outputAborted; }
        }

        public byte[] GetOutputImage(int frame)
        {
            try
            {
                // f[^z񏉊
                if (
                    _lastImageData == null ||
                    _lastImageData.Length < OutputImageFrameSize)
                {
                    _lastImageData = new byte[OutputImageFrameSize];
                }

                // 24rbgBMP(vr[ɂg)
                if (_lastImage == null || _lastImage.Size != _imageSize)
                {
                    if (_lastImage != null)
                    {
                        // 񓯊vr[ɎgĂ邽߁A null Ă
                        // Dispose ȂƕsȃANZXɂȂ邩H
                        Bitmap tmp = _lastImage;
                        _lastImage = null;
                        tmp.Dispose();
                    }
                    _lastImage = new Bitmap(
                        _imageSize.Width, _imageSize.Height,
                        PixelFormat.Format24bppRgb);
                }

                // t[ʒu`FbN
                if (frame < 0 || frame >= OutputImageFrameCount)
                {
                    throw new ArgumentOutOfRangeException(
                        "frame", frame, "t[͈͊OłB");
                }

                if (_imageFormat != null)
                {
                    // 摜f[^24rbgBMPɕ`
                    using (Bitmap orgImage = new Bitmap(GetImageFilePath(frame)))
                    using (Graphics g = Graphics.FromImage(_lastImage))
                    {
                        g.CompositingMode = CompositingMode.SourceCopy;
                        g.DrawImage(orgImage, 0, 0);
                    }

                    // f[^z֏o
                    BitmapData bmpData = _lastImage.LockBits(
                        new Rectangle(0, 0, _imageSize.Width, _imageSize.Height),
                        ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
                    Marshal.Copy(
                        bmpData.Scan0, _lastImageData, 0, OutputImageFrameSize);
                    _lastImage.UnlockBits(bmpData);
                }
                else if (_imageFileType == WorkImageType.Bgr24)
                {
                    // BGR24̓oCif[^̂܂܃f[^zɂȂ
                    using (FileStream fs = new FileStream(
                        GetImageFilePath(frame), FileMode.Open, FileAccess.Read))
                    {
                        fs.Read(_lastImageData, 0, OutputImageFrameSize);
                        fs.Close();
                    }

                    // 32t[Ƀvr[p24rbgBMP֏o
                    if ((frame & 0x1F) == 0)
                    {
                        BitmapData bmpData = _lastImage.LockBits(
                            new Rectangle(0, 0, _imageSize.Width, _imageSize.Height),
                            ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);
                        Marshal.Copy(
                            _lastImageData, 0, bmpData.Scan0, OutputImageFrameSize);
                        _lastImage.UnlockBits(bmpData);
                    }
                }
                else
                {
                    // ɂ͂Ȃ͂c
                    throw new InvalidOperationException("ԉ摜t@C`słB");
                }
            }
            catch
            {
                // [NA
                if (_lastImageData != null)
                {
                    Array.Clear(_lastImageData, 0, _lastImageData.Length);
                }
            }

            return _lastImageData;
        }

        public byte[] GetOutputAudio(int start, int length)
        {
            try
            {
                // `FbN
                if (length < 0)
                {
                    if (_lastAudioData == null)
                    {
                        _lastAudioData = new byte[0];
                    }
                    throw new ArgumentOutOfRangeException(
                        "length", length, "擾Tv 0 łB");
                }

                // f[^z񏉊
                int dataSize = length * OutputAudioSampleSize;
                if (_lastAudioData == null || _lastAudioData.Length < dataSize)
                {
                    _lastAudioData = new byte[dataSize];
                }

                // ͈̓`FbN
                if (start < 0)
                {
                    throw new ArgumentOutOfRangeException(
                        "start", start, "JnTvʒu 0 łB");
                }
                if (start + length >= OutputAudioSampleCount)
                {
                    throw new ArgumentOutOfRangeException(
                        "length", length, "擾Tv͈͈͂͊OłB");
                }

                // 1t[̃Tv擾
                int samplePerFrm = AudioSampleCountPerFrame;
                if (samplePerFrm <= 0)
                {
                    throw new InvalidOperationException(
                        "1 t[̉Tv 0 łB");
                }

                // JnEIt[ƃt[̃oCgʒu
                int beginFrm = start / samplePerFrm;
                int beginPos = (start % samplePerFrm) * OutputAudioSampleSize;
                int endFrm = (start + length) / samplePerFrm;
                int endPos =
                    ((start + length) % samplePerFrm) * OutputAudioSampleSize;

                // et[t@C擾
                int written = 0;
                for (int frame = beginFrm; frame <= endFrm; ++frame)
                {
                    if (frame == endFrm && endPos == 0) { break; }

                    // 1t[ׂēǂݍ
                    byte[] data = null;
                    try
                    {
                        data = File.ReadAllBytes(GetAudioFilePath(frame));
                    }
                    catch
                    {
                        data = new byte[
                            AudioSampleCountPerFrame * OutputAudioSampleSize];
                        Array.Clear(data, 0, data.Length);
                    }

                    // Rs[͈͂ݒ
                    int offset = 0, size = data.Length;
                    if (frame == beginFrm)
                    {
                        offset = beginPos;
                        size -= offset;
                    }
                    if (frame == endFrm)
                    {
                        size -= data.Length - endPos;
                    }

                    // Rs[
                    if (size > 0)
                    {
                        Array.Copy(
                            data, offset, _lastAudioData, written, size);
                        written += size;
                    }
                }

                // TCYȂ 0 Ŗ߂
                int rest = _lastAudioData.Length - written;
                if (rest > 0)
                {
                    Array.Clear(_lastAudioData, written, rest);
                }
            }
            catch
            {
                // [NA
                if (_lastAudioData != null)
                {
                    Array.Clear(_lastAudioData, 0, _lastAudioData.Length);
                }
            }

            return _lastAudioData;
        }

        public OutputVideoFrameFlags GetVideoFrameFlags(int frame)
        {
            // t[͂킩Ȃ̂ None Ԃ
            return OutputVideoFrameFlags.None;
        }

        public FrameProgressDelegate FrameProgress
        {
            get { return _frameProgress; }
            set { _frameProgress = value; }
        }

        public UpdatePreviewDelegate UpdatePreview
        {
            get { return _updatePreview; }
            set { _updatePreview = value; }
        }

        #endregion

        #region IHavePreviewImage o

        public Image PreviewImage
        {
            get { return _lastImage; }
        }

        #endregion

        #region IAbortable o

        public void Abort()
        {
            _outputAborted = true;
        }

        #endregion
    }
}
