using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using ruche.aviutl.plugins;

namespace ruche.nive.aviutlPluginPorts
{
    public partial class MediaInputConfigDialog : Form
    {
        /// <summary>
        /// Ōɓǂݍ񂾃vOCt@CpX̃L[B
        /// </summary>
        private static readonly string LastPluginFilePathKeyName =
            typeof(MediaInputConfigDialog).Name + ":LastPluginFilePath";

        /// <summary>
        /// fBAt@C̃tpXB
        /// </summary>
        private string _mediaFilePath = "";

        /// <summary>
        /// vOCt@C̈ꗗB
        /// </summary>
        private BindingList<FileInfo> _pluginFiles =
            new BindingList<FileInfo>();

        /// <summary>
        /// _CAO\ݒ̑IԁB
        /// </summary>
        private bool _notShown = true;

        /// <summary>
        /// fBAt@C̃tpX擾܂͐ݒ肵܂B
        /// </summary>
        public string MediaFilePath
        {
            get { return _mediaFilePath; }
            set
            {
                _mediaFilePath = Path.GetFullPath(value) ?? "";
            }
        }

        /// <summary>
        /// vOCt@C̈ꗗ擾܂͐ݒ肷B
        /// </summary>
        public IList<FileInfo> PluginFiles
        {
            get { return _pluginFiles; }
            set
            {
                _pluginFiles.Clear();
                if (value != null)
                {
                    foreach (FileInfo fi in value)
                    {
                        _pluginFiles.Add(fi);
                    }
                }
                _pluginFiles.ResetBindings();
            }
        }

        /// <summary>
        /// _CAO\ݒ̑IԂ擾B
        /// </summary>
        public bool IsNotShownChecked
        {
            get
            {
                return (_notShown && PluginFiles.Count > 0);
            }
        }

        /// <summary>
        /// RXgN^B
        /// </summary>
        public MediaInputConfigDialog()
        {
            InitializeComponent();
        }

        /// <summary>
        /// ݂̏ԂɉăRg[XVB
        /// </summary>
        private void UpdateControls()
        {
            InputPlugin plugin = null;
            bool curDirPushed = false;

            try
            {
                // vOCꗗɃvOCł邩
                bool exists = (_pluginFiles.Count > 0);
                if (exists)
                {
                    if (listPlugins.SelectedIndex < 0)
                    {
                        // IĂȂ΋IɑI
                        listPlugins.SelectedIndex = 0;

                        // I_ UpdateControls Ă΂邽߈ȍ~͕̏sv
                        return;
                    }

                    string filePath = listPlugins.SelectedValue.ToString();

                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(filePath));
                    curDirPushed = true;

                    try
                    {
                        // IꂽvOCǂݎ
                        plugin = new InputPlugin(filePath);
                    }
                    catch
                    {
                        plugin = null;
                    }
                }

                // L/ݒ
                buttonUp.Enabled =
                    (exists && listPlugins.SelectedIndex > 0);
                buttonDown.Enabled =
                    (exists && listPlugins.SelectedIndex < _pluginFiles.Count - 1);
                buttonConfig.Enabled =
                    (exists && plugin != null && plugin.HasConfig);
                buttonRemove.Enabled = exists;
                checkNotShown.Enabled = exists;
                labelNotShown.Enabled = exists;
                buttonOk.Enabled = exists;

                if (plugin != null)
                {
                    // vOCݒ
                    textPluginName.Text = plugin.PluginName;
                    textPluginDesc.Text = plugin.PluginInfo;
                    textPluginPath.Text = plugin.PluginFilePath;

                    // vOC
                    plugin.Dispose();
                    plugin = null;
                }
                else if (exists)
                {
                    textPluginName.Text = "<̓vOCł͂܂>";
                    textPluginDesc.Text = "<̓vOCł͂܂>";
                    textPluginPath.Text = listPlugins.SelectedValue.ToString();
                }
                else
                {
                    textPluginName.Text = "";
                    textPluginDesc.Text = "";
                    textPluginPath.Text = "";
                }
            }
            finally
            {
                if (curDirPushed)
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        /// <summary>
        /// _CAO[h̏B
        /// </summary>
        private void MediaInputConfigDialog_Load(object sender, EventArgs e)
        {
            // fBAt@CpXݒ
            textMedia.Text = MediaFilePath;
            textMedia.Select(textMedia.Text.Length, 0);

            // vOCꗗf[^\[Xɐݒ
            listPlugins.DataSource = _pluginFiles;
            listPlugins.DisplayMember = "Name";
            listPlugins.ValueMember = "FullName";

            // Rg[XV
            UpdateControls();

            // ͊Jn{^ɃtH[JX
            buttonOk.Select();
        }

        private void listPlugins_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Rg[XV
            UpdateControls();
        }

        private void buttonAdd_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog dlg = new OpenFileDialog())
            {
                // Ōɓǂݍ񂾃vOCt@CpXfBNgpXI
                string lastPath =
                    LibraryConfig.Datas[LastPluginFilePathKeyName] as string;
                if (lastPath == null)
                {
                    lastPath = "";
                }
                else
                {
                    try
                    {
                        lastPath = Path.GetDirectoryName(lastPath);
                        if (!Directory.Exists(lastPath))
                        {
                            lastPath = "";
                        }
                    }
                    catch
                    {
                        lastPath = "";
                    }
                }

                // _CAOݒ
                dlg.DefaultExt = "aui";
                dlg.InitialDirectory = lastPath;
                dlg.Filter =
                    "AviUtl̓vOC (*.aui)|*.aui|ׂẴt@C (*.*)|*.*";
                dlg.FilterIndex = 1;
                dlg.Title = "AviUtl̓vOCI";

                if (dlg.ShowDialog(this) == DialogResult.OK)
                {
                    try
                    {
                        // JgfBNgt@C̈ʒu
                        CurrentDirectoryStack.Instance.Push(
                            Path.GetDirectoryName(dlg.FileName));

                        // [hł邩mF
                        InputPlugin plugin = new InputPlugin(dlg.FileName);
                        plugin.Dispose();
                        plugin = null;
                    }
                    catch (Exception ex)
                    {
#if DEBUG
                        MessageBox.Show(
                            this, ex.Message + "\n\n" + ex.StackTrace,
                            ex.GetType().FullName, MessageBoxButtons.OK);
#else
                        MessageBox.Show(
                            this, ex.Message, "G[", MessageBoxButtons.OK);
#endif
                        return;
                    }
                    finally
                    {
                        CurrentDirectoryStack.Instance.Pop();
                    }

                    // t@C쐬
                    FileInfo fi = new FileInfo(dlg.FileName);

                    // vOCꗗɒǉăRg[ɔf
                    _pluginFiles.Add(fi);
                    _pluginFiles.ResetBindings();

                    // Rg[XV
                    UpdateControls();

                    // Ōɓǂݍ񂾃vOCt@CpXۑ
                    LibraryConfig.Datas[LastPluginFilePathKeyName] = fi.FullName;
                    LibraryConfig.Datas.Save();
                }
            }
        }

        private void buttonRemove_Click(object sender, EventArgs e)
        {
            if (listPlugins.SelectedIndex >= 0)
            {
                // vOCꗗ폜ăRg[ɔf
                _pluginFiles.RemoveAt(listPlugins.SelectedIndex);
                _pluginFiles.ResetBindings();

                // Rg[XV
                UpdateControls();
            }
        }

        private void buttonUp_Click(object sender, EventArgs e)
        {
            int index = listPlugins.SelectedIndex;
            if (index > 0)
            {
                // ֈړ
                FileInfo fi = _pluginFiles[index];
                _pluginFiles.RemoveAt(index);
                _pluginFiles.Insert(index - 1, fi);

                // ICfbNXXV(UpdateControls Ă΂)
                listPlugins.SelectedIndex = index - 1;
            }
        }

        private void buttonDown_Click(object sender, EventArgs e)
        {
            int index = listPlugins.SelectedIndex;
            if (index < _pluginFiles.Count)
            {
                // ֈړ
                FileInfo fi = _pluginFiles[index];
                _pluginFiles.RemoveAt(index);
                _pluginFiles.Insert(index + 1, fi);

                // ICfbNXXV(UpdateControls Ă΂)
                listPlugins.SelectedIndex = index + 1;
            }
        }

        private void buttonConfig_Click(object sender, EventArgs e)
        {
            if (listPlugins.SelectedIndex >= 0)
            {
                InputPlugin plugin = null;
                bool curDirPushed = false;

                try
                {
                    string filePath = listPlugins.SelectedValue.ToString();

                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(filePath));
                    curDirPushed = true;

                    // vOC擾
                    plugin = new InputPlugin(filePath);
                    if (!plugin.HasConfig)
                    {
                        throw new InvalidOperationException(
                            plugin.PluginFileName + " ͐ݒʂĂ܂B");
                    }

                    // ݒ
                    plugin.ShowConfig(this.Handle);
                }
                catch (Exception ex)
                {
                #if DEBUG
                    MessageBox.Show(
                        this, ex.Message + "\n\n" + ex.StackTrace,
                        ex.GetType().FullName, MessageBoxButtons.OK);
                #else
                    MessageBox.Show(
                        this, ex.Message, "G[", MessageBoxButtons.OK);
                #endif
                }
                finally
                {
                    if (plugin != null)
                    {
                        plugin.Dispose();
                        plugin = null;
                    }

                    if (curDirPushed)
                    {
                        CurrentDirectoryStack.Instance.Pop();
                    }
                }
            }
        }

        private void checkNotShown_CheckedChanged(object sender, EventArgs e)
        {
            _notShown = checkNotShown.Checked;
        }

        private void buttonOk_Click(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.OK;
        }

        private void buttonCancel_Click(object sender, EventArgs e)
        {
            this.DialogResult = DialogResult.Cancel;
        }
    }
}