using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.IO;
using ruche.aviutl.plugins;

namespace ruche.nive.aviutlPluginPorts
{
    /// <summary>
    /// AviUtlo̓vOCɂo͐ݒs_CAOB
    /// </summary>
    public partial class MediaOutputConfigDialog : Form
    {
        /// <summary>
        /// Ōɐݒ肵ԉ摜t@C^CṽL[B
        /// </summary>
        private static readonly string LastWorkImageTypeKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastWorkImageType";

        /// <summary>
        /// ŌɓǂݍAviUtlo̓vOC̃L[B
        /// </summary>
        private static readonly string LastAviUtlOutputPluginKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastAviUtlOutputPlugin";

        /// <summary>
        /// Ōɐݒ肵o͐fBNg̃L[B
        /// </summary>
        private static readonly string LastOutputDirectoryKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastOutputDirectory";

        /// <summary>
        /// Ōɐݒ肵ԃt@C폜ݒ̃L[B
        /// </summary>
        private static readonly string LastDeleteWorkKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastDeleteWork";

        /// <summary>
        /// Ōɐݒ肵YUY2XP[O^CṽL[B
        /// </summary>
        private static readonly string LastYuy2ScaleKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastYuy2Scale_2";

        /// <summary>
        /// Ōɐݒ肵YUY2Fݒ^CṽL[B
        /// </summary>
        private static readonly string LastYuy2ColorMatrixKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastYuy2ColorMatrix";

        /// <summary>
        /// Ōɐݒ肵YUY2UV̗p@̃L[B
        /// </summary>
        private static readonly string LastYuy2UVKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastYuy2UV";

        /// <summary>
        /// ԉ摜t@C̃C[W^CvB
        /// </summary>
        private WorkImageType _workImageType = WorkImageType.Png;

        /// <summary>
        /// o̓vOC̃C^tF[XB
        /// </summary>
        private OutputPlugin _plugin = null;

        /// <summary>
        /// o̓t@CpXB
        /// </summary>
        private string _fileName = null;

        /// <summary>
        /// ԃt@C폜ݒB
        /// </summary>
        private bool _deleteWork = true;

        /// <summary>
        /// ԉ摜t@C̃C[W^Cv擾B
        /// </summary>
        public WorkImageType WorkImageType
        {
            get { return _workImageType; }
        }

        /// <summary>
        /// o̓vOC̃C^tF[X擾B
        /// </summary>
        public OutputPlugin Plugin
        {
            get { return _plugin; }
        }

        /// <summary>
        /// o͂t@C̃pX擾B
        /// </summary>
        public string FileName
        {
            get { return _fileName; }
        }

        /// <summary>
        /// o͌ɒԃt@C폜邩ۂ̐ݒ擾B
        /// </summary>
        public bool DeleteWork
        {
            get { return _deleteWork; }
        }

        public MediaOutputConfigDialog()
        {
            InitializeComponent();
        }

        /// <summary>
        /// o̓vOC[hB
        /// </summary>
        /// <param name="auoFilePath">t@C̃pXB</param>
        private OutputPlugin LoadOutputPlugin(string auoFilePath)
        {
            // `FbN
            if (auoFilePath == null)
            {
                throw new ArgumentNullException("auoFilePath");
            }
            if (!File.Exists(auoFilePath))
            {
                throw new ArgumentException(
                    "w肵o̓vOCt@C݂͑܂B",
                    "auoFilePath");
            }

            OutputPlugin plugin = null;
            try
            {
                // JgfBNg auoFilePath ̏ꏊɂ
                CurrentDirectoryStack.Instance.Push(
                    Path.GetDirectoryName(auoFilePath));

                // o̓vOCC^tF[X쐬
                plugin = new OutputPlugin(auoFilePath);
            }
            finally
            {
                CurrentDirectoryStack.Instance.Pop();
            }

            return plugin;
        }

        /// <summary>
        /// o̓vOCB
        /// </summary>
        private void FreeOutputPlugin()
        {
            if (_plugin != null)
            {
                try
                {
                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    _plugin.Dispose();
                    _plugin = null;
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        /// <summary>
        /// vOCǂݍ݂ɐƂ̃Rg[XVsB
        /// </summary>
        private void UpdateControlsOnPluginLoaded()
        {
            // Rg[XV
            textPlugin.Text = _plugin.PluginInfo;
            textPlugin.Enabled = true;
            textPath.Clear();
            textPath.Enabled = true;
            buttonPath.Enabled = true;
            buttonDetail.Enabled = _plugin.HasConfig;
            comboYuy2Scale.Enabled = true;
            comboYuy2ColorMatrix.Enabled = true;
            comboYuy2UV.Enabled = true;
            buttonOk.Enabled = false;

            // t@Cݒ{^ANeBuɂ
            this.ActiveControl = buttonPath;
        }

        private void MediaOutputDialog_Load(object sender, EventArgs e)
        {
            // 
            FreeOutputPlugin();
            _fileName = null;

            // vOCI{^ANeBuɂ
            this.ActiveControl = buttonPlugin;

            // ԉ摜t@C^Cvݒ
            comboImageFormat.SelectedIndex = 0;
            if (LibraryConfig.Datas.ContainsKey(LastWorkImageTypeKeyName))
            {
                object value = LibraryConfig.Datas[LastWorkImageTypeKeyName];
                if (value is WorkImageType)
                {
                    comboImageFormat.SelectedIndex = (int)(WorkImageType)value;
                }
            }

            if (LibraryConfig.Datas.ContainsKey(LastAviUtlOutputPluginKeyName))
            {
                // OŌɓǂݍ񂾃Cu̓ǂݍݎs
                try
                {
                    string auoPath =
                        LibraryConfig.Datas[LastAviUtlOutputPluginKeyName]
                            as string;
                    if (auoPath != null)
                    {
                        // vOC[h
                        _plugin = LoadOutputPlugin(auoPath);

                        // Rg[XV
                        UpdateControlsOnPluginLoaded();
                    }
                }
                catch { }
            }

            // ԃt@C폜ݒ
            checkDeleteWork.Checked = true;
            if (LibraryConfig.Datas.ContainsKey(LastDeleteWorkKeyName))
            {
                object value = LibraryConfig.Datas[LastDeleteWorkKeyName];
                if (value is bool)
                {
                    checkDeleteWork.Checked = (bool)value;
                }
            }

            // YUY2XP[O^Cv̐ݒ
            comboYuy2Scale.SelectedIndex = (int)Yuy2ScaleType.Compressed;
            if (LibraryConfig.Datas.ContainsKey(LastYuy2ScaleKeyName))
            {
                object value = LibraryConfig.Datas[LastYuy2ScaleKeyName];
                if (value is Yuy2ScaleType)
                {
                    comboYuy2Scale.SelectedIndex = (int)(Yuy2ScaleType)value;
                }
            }

            // YUY2Fݒ^Cv̐ݒ
            comboYuy2ColorMatrix.SelectedIndex = (int)Yuy2ColorMatrixType.Auto;
            if (LibraryConfig.Datas.ContainsKey(LastYuy2ColorMatrixKeyName))
            {
                object value = LibraryConfig.Datas[LastYuy2ColorMatrixKeyName];
                if (value is Yuy2ColorMatrixType)
                {
                    comboYuy2ColorMatrix.SelectedIndex =
                        (int)(Yuy2ColorMatrixType)value;
                }
            }

            // YUY2UV̗p@̐ݒ
            comboYuy2UV.SelectedIndex = (int)Yuy2UVType.LeftOrigin;
            if (LibraryConfig.Datas.ContainsKey(LastYuy2UVKeyName))
            {
                object value = LibraryConfig.Datas[LastYuy2UVKeyName];
                if (value is Yuy2UVType)
                {
                    comboYuy2UV.SelectedIndex = (int)(Yuy2UVType)value;
                }
            }
        }

        private void MediaOutputConfigDialog_FormClosing(
            object sender, FormClosingEventArgs e)
        {
            // ԃt@C֘AXV
            LibraryConfig.Datas[LastWorkImageTypeKeyName] =
                (WorkImageType)comboImageFormat.SelectedIndex;
            LibraryConfig.Datas[LastDeleteWorkKeyName] =
                checkDeleteWork.Checked;

            if (_plugin != null)
            {
                // XV
                LibraryConfig.Datas[LastAviUtlOutputPluginKeyName] =
                    _plugin.PluginFilePath;
                LibraryConfig.Datas[LastYuy2ScaleKeyName] =
                    (Yuy2ScaleType)comboYuy2Scale.SelectedIndex;
                LibraryConfig.Datas[LastYuy2ColorMatrixKeyName] =
                    (Yuy2ColorMatrixType)comboYuy2ColorMatrix.SelectedIndex;
                LibraryConfig.Datas[LastYuy2UVKeyName] =
                    (Yuy2UVType)comboYuy2UV.SelectedIndex;
                if (_fileName != null)
                {
                    LibraryConfig.Datas[LastOutputDirectoryKeyName] =
                        Path.GetDirectoryName(_fileName);
                }

                // LZȂvOC
                if (DialogResult == DialogResult.Cancel)
                {
                    FreeOutputPlugin();
                }
            }

            // ۑ
            try
            {
                LibraryConfig.Datas.Save();
            }
            catch { }
        }

        private void buttonPlugin_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog dlg = new OpenFileDialog())
            {
                dlg.DefaultExt = "auo";
                dlg.Filter =
                    "AviUtlo̓vOC (*.auo)|*.auo|ׂẴt@C (*.*)|*.*";
                dlg.FilterIndex = 1;
                dlg.Title = "AviUtlo̓vOCI";
                if (_plugin != null)
                {
                    dlg.InitialDirectory =
                        Path.GetDirectoryName(_plugin.PluginFilePath);
                }

                if (dlg.ShowDialog(this) == DialogResult.OK)
                {
                    try
                    {
                        // vOC̃[h
                        OutputPlugin pi = LoadOutputPlugin(dlg.FileName);
                        FreeOutputPlugin();
                        _plugin = pi;

                        // Rg[XV
                        UpdateControlsOnPluginLoaded();
                    }
                    catch (Exception ex)
                    {
                    #if DEBUG
                        MessageBox.Show(
                            this, ex.Message + "\n\n" + ex.StackTrace,
                            ex.GetType().FullName, MessageBoxButtons.OK);
                    #else
                        MessageBox.Show(
                            this, ex.Message, "G[", MessageBoxButtons.OK);
                    #endif
                    }
                }
            }
        }

        private void buttonPath_Click(object sender, EventArgs e)
        {
            using (SaveFileDialog dlg = new SaveFileDialog())
            {
                dlg.Filter = _plugin.FileFilter;
                dlg.FilterIndex = 1;
                dlg.Title = _plugin.PluginName;
                if (LibraryConfig.Datas.ContainsKey(LastOutputDirectoryKeyName))
                {
                    dlg.InitialDirectory =
                        LibraryConfig.Datas[LastOutputDirectoryKeyName] as string;
                }

                if (dlg.ShowDialog(this) == DialogResult.OK)
                {
                    // Rg[XV
                    textPath.Text = dlg.FileName;
                    buttonOk.Enabled = true;

                    // Jn{^ANeBuɂ
                    this.ActiveControl = buttonOk;
                }
            }
        }

        private void buttonDetail_Click(object sender, EventArgs e)
        {
            if (_plugin.HasConfig)
            {
                try
                {
                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    // ݒ_CAOĂяo
                    _plugin.ShowConfig(this.Handle);
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        private void buttonOk_Click(object sender, EventArgs e)
        {
            // ԃt@C^Cvݒ
            _workImageType = (WorkImageType)comboImageFormat.SelectedIndex;

            // t@CpXݒ
            _fileName = textPath.Text;

            // ԃt@C폜ݒ
            _deleteWork = checkDeleteWork.Checked;

            // vOCYUY֘Aݒ
            _plugin.Yuy2Scale = (Yuy2ScaleType)comboYuy2Scale.SelectedIndex;
            _plugin.Yuy2ColorMatrix =
                (Yuy2ColorMatrixType)comboYuy2ColorMatrix.SelectedIndex;
            _plugin.Yuy2UV = (Yuy2UVType)comboYuy2UV.SelectedIndex;

            DialogResult = DialogResult.OK;
        }
    }
}